       >>SOURCE FORMAT IS FIXED
      ******************************************************************
      * Author: Jay Moseley
      * Date: July 19, 2020
      * Purpose: Read Hercules Emulated Tape (.het) file and produce
      *          content report and optionally dump contents to print.
      * Tectonics: cobc -x tapedump.cbl 
      *----------------------------------------------------------------*
      *                                                                *
      * 06 April 2021 - GnuCOBOL 3.1.2.0 introduced a fix for data     *
      *                 items defined with ANY LENGTH, which rendered  *
      *                 compile errors to coding methodolgy I had been *
      *                 using. Revision made to this code [subprogram  *
      *                 AtoEtoA] to comply with the fix implemented.   *
      ******************************************************************
       IDENTIFICATION DIVISION.
       PROGRAM-ID. TAPEDUMP.
       AUTHOR.
       DATE-WRITTEN. JULY 19, 2020.
       DATE-COMPILED.

       REMARKS. This program is placed in the public domain and may be
                freely used and incorporated into derived works so
                long as attribution to the original authorship remains 
                in any distributed copies of the COBOL source.

       ENVIRONMENT DIVISION.
       CONFIGURATION SECTION.
       REPOSITORY.
           FUNCTION ALL INTRINSIC.

       INPUT-OUTPUT SECTION.
       FILE-CONTROL.

       I-O-CONTROL.

       DATA DIVISION.
       FILE SECTION.

       WORKING-STORAGE SECTION.

       *> Arguments passed from command line
       01  WS-ARGUMENT-COUNT           PIC 9(02).
       01  WS-ARGUMENT                 OCCURS 4 TIMES
                                       PIC X(64).
           88  ARGUMENT-IS-INPUT       VALUE 'input', 'INPUT',
                                             'in', 'IN',
                                             'i', 'I',
                                             'input:', 'INPUT:',
                                             'in:', 'IN:',
                                             'i:', 'I:'.
           88  ARGUMENT-IS-DETAIL      VALUE 'detail', 'DETAIL',
                                             'd', 'D'.
       01  WS-ARGUMENT-INPUT-SWITCH    PIC X VALUE 'N'.
           88  ARGUMENT-INPUT-GIVEN          VALUE 'Y'.
           88  ARGUMENT-INPUT-NOT-GIVEN      VALUE 'N'.
       01  WS-ARGUMENT-DETAIL-SWITCH   PIC X VALUE 'S'.
           88  ARGUMENT-DETAIL-GIVEN         VALUE 'D'.
           88  ARGUMENT-DETAIL-NOT-GIVEN     VALUE 'S'.

       *> Parameters for CBL_ file handling routines.
       01  CBL-PARMS.
           02  CBL-INPUT-FILE          PIC X(64) VALUE SPACES.
           02  CBL-END-FILE-SWITCH     PIC X VALUE 'N'.
               88  CBL-END-OF-FILE           VALUE 'Y'.
               88  CBL-NOT-END-OF-FILE       VALUE 'N'.
           02  CBL-ACCESS-MODE         PIC X COMP-X VALUE 1.
           02  CBL-HANDLE              PIC X(4) COMP-X.
           02  CBL-OFFSET              PIC X(8) COMP-X VALUE 0.
           02  CBL-NBYTES              PIC X(4) COMP-X.
           02  CBL-BUFFER              PIC X(32768).

       *> Last header record read
       01  HEADER-RECORD.
           02  HDR-CURRENT-LRECL       PIC S9(4) COMP-5.
           02  HDR-PREVIOUS-LRECL      PIC S9(4) COMP-5.
           02  HDR-FLAGS               PIC S9(4) COMP-5.

       *> Label records information (from MVS header/trailer records)
       01  MVS-LABEL-INFO.
           02  ML-FILE-NAME            PIC X(17).
           02  ML-FILE-NUMBER          PIC 9(4).
           02  ML-RECFM                PIC X.
           02  ML-BLKSIZE              PIC 9(5).
           02  ML-BLKSIZE-EDIT         PIC ZZ,ZZ9.
           02  ML-LRECL                PIC 9(5).
           02  ML-LRECL-EDIT           PIC ZZ,ZZ9.
           02  ML-BLOCK-COUNT          PIC 9(6).
           02  ML-BLOCK-COUNT-EDIT     PIC ZZZ,ZZ9.
           02  ML-AUDIT-ERROR.
               03  FILLER              PIC X(18) VALUE
                   '  *** [From EOF1: '.
               03  ML-AUDIT-COUNT      PIC ZZZ,ZZ9.
               03  FILLER              PIC X(1) VALUE ']'.

       *> Various work fields
       01  FIRST-HEADER-SWITCH         PIC X VALUE 'Y'.
           88  FIRST-HEADER                  VALUE 'Y'.
           88  NOT-FIRST-HEADER              VALUE 'N'.
       01  IX1                         PIC S9(4) COMP.
       01  IX2                         PIC S9(4) COMP.
       01  OFFSET                      PIC S9(5) COMP.
       01  FLAG-REM                    PIC 9.
       01  FLAG-VALUES.
           02 FILLER                   PIC X(6) VALUE 'X''80'''.
           02 FLAG-40                  PIC X(6) VALUE 'X''40'''.
           02 FLAG-20                  PIC X(6) VALUE 'X''20'''.
           02 FLAG-10                  PIC X(6) VALUE 'X''10'''.
       01  FLAG-VALUE                  REDEFINES FLAG-VALUES
                                       OCCURS 4 TIMES
                                       PIC X(6).
       01  AUDIT-BLOCK-COUNT           PIC S9(4).

       *> Display block for header
       01  HEADER-DISPLAY-BLOCK.
           02  FILLER                  PIC X(09) VALUE '[Header] '.
           02  FILLER                  PIC X(17) VALUE 
               'This Block Size: '.
           02  HDB-THIS-BLKSIZE        PIC ZZ,ZZ9.
           02  FILLER                  PIC X(02) VALUE SPACES.
           02  FILLER                  PIC X(17) VALUE
               'Prev Block Size: '.
           02  HDB-PREV-BLKSIZE        PIC ZZ,ZZ9.
           02  FILLER                  PIC X(02) VALUE SPACES.
           02  FILLER                  PIC X(07) VALUE 'Flags: '.
           02  HDB-FLAGS               PIC X(24).

       *> File information heading
       01  FILE-INFO-HEADER.
           02  FILLER                  PIC X(06) VALUE 'File  '.
           02  FILLER                  PIC X(11) VALUE 'MVS Dataset'.
           02  FILLER                  PIC X(07) VALUE SPACES.
           02  FILLER                  PIC X(06) VALUE 'RECFM '.
           02  FILLER                  PIC X(09) VALUE 'BLKSIZE  '.
           02  FILLER                  PIC X(07) VALUE 'LRECL  '.
           02  FILLER                  PIC X(11) VALUE 'Blocks Read'.

       PROCEDURE DIVISION.

       HOUSEKEEPING.

      * ************************************************************* *
      * Retrieve ARGUMENT count. If no arguments, display syntax and  *
      * exit; otherwise, retrieve arguments and evaluate them.        *
      * ************************************************************* *
           ACCEPT WS-ARGUMENT-COUNT FROM ARGUMENT-NUMBER.
           IF WS-ARGUMENT-COUNT = 0
               DISPLAY 'Syntax: tapedump input: | in: | i: '
                       '<input file name>'
               DISPLAY '                 [ detail | d ]'
               END-DISPLAY
               STOP RUN RETURNING 101
           END-IF.
      *> Limit arguments to 3
           IF WS-ARGUMENT-COUNT > 3
             MOVE 3 TO WS-ARGUMENT-COUNT
           END-IF.
           PERFORM VARYING IX1
                   FROM 1 BY 1
                   UNTIL IX1 > WS-ARGUMENT-COUNT
               DISPLAY IX1 UPON ARGUMENT-NUMBER
               ACCEPT WS-ARGUMENT (IX1) FROM ARGUMENT-VALUE
           END-PERFORM.
           PERFORM VARYING IX1
                   FROM 1 BY 1
                   UNTIL IX1 > WS-ARGUMENT-COUNT
             IF ARGUMENT-IS-INPUT (IX1)
               IF WS-ARGUMENT-COUNT = IX1
                 DISPLAY 'Input file name not specified'
                 STOP RUN RETURNING 102
               ELSE
                 SET ARGUMENT-INPUT-GIVEN TO TRUE
                 COMPUTE IX2 = IX1 + 1
                 MOVE WS-ARGUMENT (IX2) TO CBL-INPUT-FILE
               END-IF
             END-IF
             IF ARGUMENT-IS-DETAIL (IX1)
               SET ARGUMENT-DETAIL-GIVEN TO TRUE
             END-IF
           END-PERFORM.

       PROCESS-INPUT-FILE.

      * ************************************************************* *
      * Process file with low-level I/O routines                      *
      * ************************************************************* *

           CALL 'CBL_OPEN_FILE' USING CBL-INPUT-FILE, CBL-ACCESS-MODE,
                                      0, 0, CBL-HANDLE.
           IF RETURN-CODE NOT = ZERO
             DISPLAY 'Non-zero return code on CBL_OPEN_FILE ('
                     RETURN-CODE ')'
             DISPLAY 'File name: ' TRIM(CBL-INPUT-FILE)
             END-DISPLAY
             STOP RUN 99
           END-IF.

           PERFORM FOREVER

             PERFORM READ-HEADER

             IF CBL-END-OF-FILE
               EXIT PERFORM
             END-IF

             IF HDR-CURRENT-LRECL > 0
               PERFORM READ-DATA-BLOCK
               PERFORM PROCESS-DATA-BLOCK
             END-IF

             IF HDR-FLAGS = 64 
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY 'Tape Mark'
               END-IF
             END-IF

           END-PERFORM.

           CALL 'CBL_CLOSE_FILE' USING CBL-HANDLE.
           GOBACK.
      * - - - - - - - - - - - - - - - - - - - - - - - Program Exit Point

       READ-HEADER.

      * ************************************************************* *
      * Each logical block on the tape is preceeded by a 6 byte       *
      * header record that contains current/previous record length    *
      * and a set of indicator 'flags' encoded into an integer field  *
      * ************************************************************* *
           
           MOVE +6 TO CBL-NBYTES.
           PERFORM CALL-CBL-READ.
           MOVE CBL-BUFFER (1:6) TO HEADER-RECORD.

           IF FIRST-HEADER
             IF HDR-CURRENT-LRECL = 80 
            AND HDR-PREVIOUS-LRECL = 0
            AND HDR-FLAGS = 160
              SET NOT-FIRST-HEADER TO TRUE
           ELSE
              DISPLAY 'Data for first header block seems '
                      'to indicate that the file'
              DISPLAY 'specified for input (' 
                      TRIM(CBL-INPUT-FILE)
                      ' is not an AWSTAPE dataset.'
              END-DISPLAY
              STOP RUN 99
           END-IF.

           IF CBL-NOT-END-OF-FILE
          AND ARGUMENT-DETAIL-GIVEN
             PERFORM SHOW-HEADER-DETAILS
           END-IF.
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point

       SHOW-HEADER-DETAILS.

      * ************************************************************* *
      * Display header block field contents                           *
      * ************************************************************* *

           MOVE SPACES TO HDB-FLAGS.
           MOVE +4 TO IX1.
           DIVIDE HDR-FLAGS BY 16 GIVING TALLY.
           PERFORM UNTIL TALLY < 1
             DIVIDE TALLY BY 2 GIVING TALLY
                                   REMAINDER FLAG-REM
             IF FLAG-REM = 1
               COMPUTE IX2 = ((IX1 - 1) * 6) + 1
               MOVE FLAG-VALUE (IX1) TO HDB-FLAGS (IX2:6)
             END-IF
             SUBTRACT 1 FROM IX1
           END-PERFORM.
           MOVE HDR-CURRENT-LRECL TO HDB-THIS-BLKSIZE.
           MOVE HDR-PREVIOUS-LRECL TO HDB-PREV-BLKSIZE.
           DISPLAY HEADER-DISPLAY-BLOCK.
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point

       READ-DATA-BLOCK.

      * ************************************************************* *
      * Using length from header, read data block from file and       *
      * translate from EBCDIC to ASCII                                *
      * ************************************************************* *
           
           MOVE HDR-CURRENT-LRECL TO CBL-NBYTES.
           PERFORM CALL-CBL-READ.
           CALL 'TOASCII' USING CBL-BUFFER (1:HDR-CURRENT-LRECL).
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point

       PROCESS-DATA-BLOCK.

      * ************************************************************* *
      * Process data block with handling for MVS control records      *
      * VOL1, HDR1, HDR2, EOF1 and EOF2, as well as individual        *
      * blocks from the dataset(s) contained on the tape              *
      * ************************************************************* *

           EVALUATE TRUE
 
             WHEN CBL-BUFFER (1:4) = 'VOL1'
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
               ELSE
                 DISPLAY 'Serial Number from VOLume 1 label: '
                       CBL-BUFFER (5:6)
                 DISPLAY FILE-INFO-HEADER
                 END-DISPLAY
               END-IF

             WHEN CBL-BUFFER (1:4) = 'HDR1'
               MOVE +0 TO AUDIT-BLOCK-COUNT
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
               END-IF
               MOVE CBL-BUFFER (5:17) TO ML-FILE-NAME
               MOVE CBL-BUFFER (32:4) TO ML-FILE-NUMBER

             WHEN CBL-BUFFER (1:4) = 'HDR2'
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
               END-IF
               MOVE CBL-BUFFER (39:1) TO ML-RECFM
               MOVE NUMVAL(CBL-BUFFER (6:5)) 
                 TO ML-BLKSIZE, ML-BLKSIZE-EDIT
               MOVE NUMVAL(CBL-BUFFER (11:5)) 
                 TO ML-LRECL, ML-LRECL-EDIT

             WHEN CBL-BUFFER (1:4) = 'EOF1'
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
               END-IF
               MOVE NUMVAL(CBL-BUFFER (55:6)) 
                 TO ML-BLOCK-COUNT,
                    ML-BLOCK-COUNT-EDIT

             WHEN CBL-BUFFER (1:4) = 'EOF2'
               IF ARGUMENT-DETAIL-GIVEN
                 DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
               ELSE
                 DISPLAY ML-FILE-NUMBER '  '
                         ML-FILE-NAME '   '
                         ML-RECFM '    '
                         ML-BLKSIZE-EDIT ' '
                         ML-LRECL-EDIT '      '
                         ML-BLOCK-COUNT-EDIT
                         WITH NO ADVANCING
                 END-DISPLAY
                 IF AUDIT-BLOCK-COUNT NOT = ML-BLOCK-COUNT
                   MOVE AUDIT-BLOCK-COUNT TO ML-AUDIT-COUNT
                   DISPLAY ML-AUDIT-ERROR
                 ELSE
                   DISPLAY ' '
                 END-IF
               END-IF

             WHEN OTHER
               ADD +1 TO AUDIT-BLOCK-COUNT
               IF ARGUMENT-DETAIL-GIVEN
                 PERFORM DISPLAY-DATA-BLOCK
               END-IF

           END-EVALUATE.
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point


       DISPLAY-DATA-BLOCK.

      * ************************************************************* *
      * Special handling required for 'blocked' records, which are    *
      * displayed 'unblocked' using LRECL from HDR1 record            *
      * ************************************************************* *

           IF ML-LRECL = ML-BLKSIZE
             DISPLAY CBL-BUFFER (1:HDR-CURRENT-LRECL)
           ELSE
             MOVE +1 TO OFFSET
             PERFORM WITH TEST AFTER 
                     UNTIL OFFSET > HDR-CURRENT-LRECL
               DISPLAY CBL-BUFFER (OFFSET:ML-LRECL)
               ADD ML-LRECL TO OFFSET
             END-PERFORM
           END-IF.
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point

       CALL-CBL-READ.

      * ************************************************************* *
      * All data is read from the file here. Number of bytes to read  *
      * each time the paragraph is performed is in CBL-NBYTES and the *
      * 'read point' in the file (CBL-OFFSET) is updated after each   *
      * successful read. Any unexpected return-code abends program    *
      * ************************************************************* *

           CALL 'CBL_READ_FILE' USING CBL-HANDLE, CBL-OFFSET,
                                      CBL-NBYTES, 0, CBL-BUFFER.
           EVALUATE TRUE
             WHEN RETURN-CODE = 0
               ADD CBL-NBYTES TO CBL-OFFSET
               CONTINUE
             WHEN RETURN-CODE = 10
               SET CBL-END-OF-FILE TO TRUE
               CONTINUE
             WHEN OTHER
               DISPLAY 'Non-zero return code on CBL_READ_FILE ('
                       RETURN-CODE ')'
               END-DISPLAY
               STOP RUN 99
           END-EVALUATE.
      * - - - - - - - - - - - - - - - - - - - - - - - Perform Exit Point

       END PROGRAM TAPEDUMP.

      ******************************************************************
      * Author: Jay Moseley
      * Date: July 19, 2020
      * Purpose: Convert passed record from EBCDIC to ASCII.
      * Tectonics: cobc -m AtoEtoA
      *----------------------------------------------------------------*
      *                                                                *
      * 06 April 2021 - GnuCOBOL 3.1.2.0 introduced a fix for data     *
      *                 items defined with ANY LENGTH, which rendered  *
      *                 compile errors to coding methodolgy I had been *
      *                 using. Revision made to this code to comply    *
      *                 with the fix implemented.                      *
      ******************************************************************
       IDENTIFICATION DIVISION.
       PROGRAM-ID. AtoEtoA.

       ENVIRONMENT DIVISION.

       CONFIGURATION SECTION.
       REPOSITORY.
           FUNCTION ALL INTRINSIC.

       DATA DIVISION.

       WORKING-STORAGE SECTION.

       *> EBCDIC to ASCII:
       01  EBCDIC-DATA.
           02  PIC X(16) VALUE X'000102030405060708090A0B0C0D0E0F'.      000-015
           02  PIC X(16) VALUE X'101112131415161718191A1B1C1D1E1F'.      016-031
           02  PIC X(16) VALUE X'202122232425262728292A2B2C2D2E2F'.      032-047
           02  PIC X(16) VALUE X'303132333435363738393A3B3C3D3E3F'.      048-063
           02  PIC X(16) VALUE X'404142434445464748494A4B4C4D4E4F'.      064-079
           02  PIC X(16) VALUE X'505152535455565758595A5B5C5D5E5F'.      080-095
           02  PIC X(16) VALUE X'606162636465666768696A6B6C6D6E6F'.      096-111
           02  PIC X(16) VALUE X'707172737475767778797A7B7C7D7E7F'.      112-127
           02  PIC X(16) VALUE X'808182838485868788898A8B8C8D8E8F'.      128-143
           02  PIC X(16) VALUE X'909192939495969798999A9B9C9D9E9F'.      144-159
           02  PIC X(16) VALUE X'A0A1A2A3A4A5A6A7A8A9AAABACADAEAF'.      160-175
           02  PIC X(16) VALUE X'B0B1B2B3B4B5B6B7B8B9BABBBCBDBEBF'.      176-191
           02  PIC X(16) VALUE X'C0C1C2C3C4C5C6C7C8C9CACBCCCDCECF'.      192-207
           02  PIC X(16) VALUE X'D0D1D2D3D4D5D6D7D8D9DADBDCDDDEDF'.      208-223
           02  PIC X(16) VALUE X'E0E1E2E3E4E5E6E7E8E9EAEBECEDEEEF'.      224-239
           02  PIC X(16) VALUE X'F0F1F2F3F4F5F6F7F8F9FAFBFCFDFEFF'.      240-255
       01  EBCDIC-ENCODING REDEFINES EBCDIC-DATA PIC X(256).

       *> ASCII to EBCDIC:
       01  ASCII-DATA.
           02  PIC X(16) VALUE X'000102039C09867F978D8E0B0C0D0E0F'.      000-015
           02  PIC X(16) VALUE X'101112139D8508871819928F1C1D1E1F'.      016-031
           02  PIC X(16) VALUE X'80818283840A171B88898A8B8C050607'.      032-047
           02  PIC X(16) VALUE X'909116939495960498999A9B14159E1A'.      048-063
           02  PIC X(16) VALUE X'20A0E2E4E0E1E3E5E7F1A22E3C282B7C'.      064-079
           02  PIC X(16) VALUE X'26E9EAEBE8EDEEEFECDF21242A293BAC'.      080-095
           02  PIC X(16) VALUE X'2D2FC2C4C0C1C3C5C7D1A62C255F3E3F'.      096-111
           02  PIC X(16) VALUE X'F8C9CACBC8CDCECFCC603A2340273D22'.      112-127
           02  PIC X(16) VALUE X'D8616263646566676869ABBBF0FDFEB1'.      128-143
           02  PIC X(16) VALUE X'B06A6B6C6D6E6F707172AABAE6B8C6A4'.      144-159
           02  PIC X(16) VALUE X'B57E737475767778797AA1BFD0DDDEAE'.      160-175
           02  PIC X(16) VALUE X'5EA3A5B7A9A7B6BCBDBE5B5DAFA8B4D7'.      176-191
           02  PIC X(16) VALUE X'7B414243444546474849ADF4F6F2F3F5'.      192-207
           02  PIC X(16) VALUE X'7D4A4B4C4D4E4F505152B9FBFCF9FAFF'.      208-223
           02  PIC X(16) VALUE X'5CF7535455565758595AB2D4D6D2D3D5'.      224-239
           02  PIC X(16) VALUE X'30313233343536373839B3DBDCD9DA9F'.      240-255
       01  ASCII-ENCODING REDEFINES ASCII-DATA PIC X(256).

       LINKAGE SECTION.

       *> Buffer from caller, containing string:
       01  PASSED-BUFFER               PIC X ANY LENGTH.

       PROCEDURE DIVISION USING PASSED-BUFFER.

       MAIN-PROCEDURE.

       ENTRY 'TOASCII' USING PASSED-BUFFER.

           TRANSFORM PASSED-BUFFER
             FROM EBCDIC-ENCODING TO ASCII-ENCODING.

           GOBACK.    

       ENTRY 'TOEBCDIC' USING PASSED-BUFFER.

           TRANSFORM PASSED-BUFFER
             FROM ASCII-ENCODING TO EBCDIC-ENCODING.

           GOBACK.    

       END PROGRAM AtoEtoA.
